import unittest

from tala.ddd.ddd_component_manager import DDDComponentManager
from tala.model.semantic_logic import SemanticLogic
from tala.model.test.semanticstester import SemanticsTester
from tala.model.ontology import Ontology
from tala.model.polarity import Polarity
from tala.model.predicate import Predicate
from tala.model.proposition import PredicateProposition, PropositionSet
from tala.model.sort import CustomSort


class SemanticLogicTests(SemanticsTester):
    def setUp(self):
        SemanticsTester.setUp(self)
        self.declare_valid_relations()
        self.declare_provoker_objects()

    def declare_valid_relations(self):
        self.declare_combinable_and_relevant([
            # WH questions
            ("?X.dest_city(X)", "dest_city(paris)", "dest_city(paris)"),
            ("?X.dest_city(X)", "dest_city(london)", "dest_city(london)"),
            ("?X.dest_city(X)", "~dest_city(paris)", "~dest_city(paris)"),
            ("?X.dest_city(X)", "~dest_city(london)", "~dest_city(london)"),
            ("?X.dest_city(X)", "paris", "dest_city(paris)"),
            ("?X.dest_city(X)", "london", "dest_city(london)"),
            ("?X.dest_city(X)", "~paris", "~dest_city(paris)"),
            ("?X.dest_city(X)", "~london", "~dest_city(london)"),
            (
                "?X.dest_city(X)", "~set([dest_city(paris), dest_city(london)])",
                "~set([dest_city(paris), dest_city(london)])"
            ),
            ("?X.goal(X)", "goal(perform(make_reservation))", "goal(perform(make_reservation))"),
            ("?X.goal(X)", "goal(perform(make_domestic_reservation))", "goal(perform(make_domestic_reservation))"),
            ("?X.goal(X)", "goal(resolve(?X.price(X)))", "goal(resolve(?X.price(X)))"),

            # Yes-no questions
            ("?dest_city(paris)", "yes", "dest_city(paris)"),
            ("?dest_city(paris)", "no", "~dest_city(paris)"),
            ("?dest_city(paris)", "~dest_city(paris)", "~dest_city(paris)"),
            ("?dest_city(paris)", "paris", "dest_city(paris)"),
            ("?dest_city(paris)", "~paris", "~dest_city(paris)"),
            ("?dest_city(paris)", "dest_city(paris)", "dest_city(paris)"),
            ("?goal(perform(make_reservation))", "yes", "goal(perform(make_reservation))"),
            ("?goal(perform(make_reservation))", "no", "~goal(perform(make_reservation))"),
            ("?goal(perform(make_reservation))", "goal(perform(make_reservation))", "goal(perform(make_reservation))"),

            # understanding questions
            ("?und(USR, dest_city(paris))", "yes", "und(USR, dest_city(paris))"),
            ("?und(USR, dest_city(paris))", "no", "~und(USR, dest_city(paris))"),
            ("?und(USR, goal(perform(make_reservation)))", "yes", "und(USR, goal(perform(make_reservation)))"),
            ("?und(USR, goal(perform(make_reservation)))", "no", "~und(USR, goal(perform(make_reservation)))"),

            # multi-alt questions
            ("?set([dest_city(paris), dest_city(london)])", "dest_city(paris)", "dest_city(paris)"),
            ("?set([dest_city(paris), dest_city(london)])", "dest_city(london)", "dest_city(london)"),
            ("?set([dest_city(paris), dest_city(london)])", "~dest_city(paris)", "~dest_city(paris)"),
            ("?set([dest_city(paris), dest_city(london)])", "~dest_city(london)", "~dest_city(london)"),
            ("?set([dest_city(paris), dest_city(london)])", "paris", "dest_city(paris)"),
            ("?set([dest_city(paris), dest_city(london)])", "london", "dest_city(london)"),
            ("?set([dest_city(paris), dest_city(london)])", "~paris", "~dest_city(paris)"),
            ("?set([dest_city(paris), dest_city(london)])", "~london", "~dest_city(london)"),
            ("?set([dest_city(paris), dest_city(london)])", "no", "~set([dest_city(paris), dest_city(london)])"),
            (
                "?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])",
                "goal(perform(make_reservation))", "goal(perform(make_reservation))"
            ),
            (
                "?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])", "goal(resolve(?X.price(X)))",
                "goal(resolve(?X.price(X)))"
            ),
            (
                "?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])", "no",
                "~set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])"
            ),

            # single-alt questions
            ("?set([dest_city(paris)])", "dest_city(paris)", "dest_city(paris)"),
            ("?set([dest_city(paris)])", "~dest_city(paris)", "~dest_city(paris)"),
            ("?set([dest_city(paris)])", "paris", "dest_city(paris)"),
            ("?set([dest_city(paris)])", "~paris", "~dest_city(paris)"),
            ("?set([dest_city(paris)])", "yes", "dest_city(paris)"),
            ("?set([dest_city(paris)])", "no", "~dest_city(paris)"),
            (
                "?set([goal(perform(make_reservation))])", "goal(perform(make_reservation))",
                "goal(perform(make_reservation))"
            ),
            ("?set([goal(perform(make_reservation))])", "yes", "goal(perform(make_reservation))"),
            ("?set([goal(perform(make_reservation))])", "no", "~goal(perform(make_reservation))"),

            # preconfirmation question
            ("?preconfirmed(MakeReservation, [])", "yes", "preconfirmed(MakeReservation, [])"),
            ("?preconfirmed(MakeReservation, [])", "no", "~preconfirmed(MakeReservation, [])"),
            # features
            ("?X.dest_city(X)", "resort", "dest_city_type(resort)"),
            ("?X.dest_city(X)", "dest_city_type(resort)", "dest_city_type(resort)"),
            ("?X.dest_city(X)", "~resort", "~dest_city_type(resort)"),
            ("?X.dest_city(X)", "~dest_city_type(resort)", "~dest_city_type(resort)"),

            # knowledge precondition question
            ("?know_answer(?X.dest_city_type(X))", "yes", "know_answer(?X.dest_city_type(X))"),
            ("?know_answer(?X.dest_city_type(X))", "no", "~know_answer(?X.dest_city_type(X))"),
            ("?know_answer(?X.dest_city_type(X))", "dest_city_type(resort)", "dest_city_type(resort)"),
            ("?know_answer(?X.dest_city_type(X))", "resort", "dest_city_type(resort)"),
            ("?know_answer(?X.dest_city_type(X))", "~dest_city_type(resort)", "~dest_city_type(resort)"),
            ("?know_answer(?X.dest_city_type(X))", "~resort", "~dest_city_type(resort)"),

            # consequent question
            (
                "?X.implies(dest_city(paris), price(X))", "implies(dest_city(paris), price(123.0))",
                "implies(dest_city(paris), price(123.0))"
            ),
        ])

        self.declare_relevant_not_resolving([
            # WH questions
            ("?X.dest_city(X)", "~set([dest_city(paris), dest_city(london)])"),

            # alt questions
            ("?set([dest_city(paris), dest_city(london)])", "no"),
            ("?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])", "no"),

            # features
            ("?X.dest_city(X)", "~dest_city_type(resort)"),
            ("?X.dest_city(X)", "~resort"),
            ("?dest_city(paris)", "resort"),
            ("?dest_city(paris)", "~resort"),
            ("?dest_city(paris)", "~dest_city_type(resort)"),
            ("?set([dest_city(paris), dest_city(london)])", "resort"),
            ("?set([dest_city(paris)])", "resort"),
            ("?set([dest_city(paris)])", "~resort"),
            ("?set([dest_city(paris)])", "~dest_city_type(resort)"),
            ("?dest_city(paris)", "dest_city_type(resort)"),
            ("?set([dest_city(paris), dest_city(london)])", "dest_city_type(resort)"),
            ("?set([dest_city(paris), dest_city(london)])", "~dest_city_type(resort)"),
            ("?set([dest_city(paris), dest_city(london)])", "~resort"),
            ("?set([dest_city(paris)])", "dest_city_type(resort)"),

            # knowledge precondition question
            ("?know_answer(?X.dest_city_type(X))", "~dest_city_type(resort)"),
            ("?know_answer(?X.dest_city_type(X))", "~resort"),
        ])

        self.declare_resolving([
            # WH questions
            ("?X.dest_city(X)", "dest_city(paris)"),
            ("?X.dest_city(X)", "dest_city(london)"),
            ("?X.dest_city(X)", "paris"),
            ("?X.dest_city(X)", "london"),
            ("?X.goal(X)", "goal(perform(make_reservation))"),
            ("?X.goal(X)", "goal(perform(make_domestic_reservation))"),
            ("?X.goal(X)", "goal(resolve(?X.price(X)))"),

            # Yes-no questions
            ("?dest_city(paris)", "dest_city(paris)"),
            ("?dest_city(paris)", "~dest_city(paris)"),
            ("?dest_city(paris)", "paris"),
            ("?dest_city(paris)", "~paris"),
            ("?dest_city(paris)", "yes"),
            ("?dest_city(paris)", "no"),
            ("?goal(perform(make_reservation))", "goal(perform(make_reservation))"),
            ("?goal(perform(make_reservation))", "yes"),
            ("?goal(perform(make_reservation))", "no"),

            # Understanding questions
            ("?und(USR, dest_city(paris))", "yes"),
            ("?und(USR, dest_city(paris))", "no"),
            ("?und(USR, goal(perform(make_reservation)))", "yes"),
            ("?und(USR, goal(perform(make_reservation)))", "no"),

            # Multi-alt questions
            ("?set([dest_city(paris), dest_city(london)])", "dest_city(paris)"),
            ("?set([dest_city(paris), dest_city(london)])", "dest_city(london)"),
            ("?set([dest_city(paris), dest_city(london)])", "paris"),
            ("?set([dest_city(paris), dest_city(london)])", "london"),
            ("?set([dest_city(paris), dest_city(london)])", "no"),
            ("?set([dest_city(paris), dest_city(london)])", "~set([dest_city(paris), dest_city(london)])"),
            ("?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])", "goal(perform(make_reservation))"),
            ("?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])", "goal(resolve(?X.price(X)))"),
            ("?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])", "no"),
            (
                "?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])",
                "~set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])"
            ),

            # single-alt questions
            ("?set([dest_city(paris)])", "dest_city(paris)"),
            ("?set([dest_city(paris)])", "~dest_city(paris)"),
            ("?set([dest_city(paris)])", "paris"),
            ("?set([dest_city(paris)])", "~paris"),
            ("?set([dest_city(paris)])", "yes"),
            ("?set([dest_city(paris)])", "no"),
            ("?set([goal(perform(make_reservation))])", "goal(perform(make_reservation))"),
            ("?set([goal(perform(make_reservation))])", "yes"),
            ("?set([goal(perform(make_reservation))])", "no"),

            # preconfirmation question
            ("?preconfirmed(MakeReservation, [])", "yes"),
            ("?preconfirmed(MakeReservation, [])", "no"),

            # special case: issues-alt-question, issue answer not in alts
            # NOTE: does not seem to be needed.
            # ("?set([goal(resolve(?X.dest_city(X))), goal(resolve(?X.dept_city(X)))])",
            #  "goal(resolve(?X.price(X)))"),
            # ("?set([goal(resolve(?X.dest_city(X))), goal(resolve(?X.dept_city(X)))])",
            #  "no"),

            # special case: answer action is dominated by action in alts
            ("?set([goal(perform(make_reservation))])", "goal(perform(make_domestic_reservation))"),
            ("?set([goal(perform(make_reservation))])", "goal(resolve(?X.price(X)))"),
            (
                "?set([goal(perform(make_reservation)), goal(resolve(?X.price(X)))])",
                "goal(perform(make_domestic_reservation))"
            ),

            # knowledge precondition question
            ("?know_answer(?X.dest_city_type(X))", "dest_city_type(resort)"),
            ("?know_answer(?X.dest_city_type(X))", "resort"),
            ("?know_answer(?X.dest_city_type(X))", "yes"),
            ("?know_answer(?X.dest_city_type(X))", "no"),
            ("?know_answer(?X.dest_city_type(X))", "~resort"),  # disputable
            ("?know_answer(?X.dest_city_type(X))", "~dest_city_type(resort)"),  # disputable

            # consequent question
            ("?X.implies(dest_city(paris), price(X))", "implies(dest_city(paris), price(123.0))"),
        ])

        self.declare_incompatible([
            ("dest_city(paris)", "dest_city(london)"),
            ("dest_city(paris)", "~dest_city(paris)"),
            ("dest_city(london)", "~dest_city(london)"),
            ("means_of_transport(plane)", "~means_of_transport(plane)"),

            # features
            ("dest_city_type(resort)", "~dest_city(london)"),
            ("dest_city_type(resort)", "~dest_city(paris)"),
            ("~dest_city(london)", "~dest_city_type(resort)"),
            ("~dest_city(paris)", "~dest_city_type(resort)"),
            ("~dest_city_type(resort)", "dest_city_type(resort)"),
        ])

    def declare_provoker_objects(self):
        self.declare_answer("means_of_transport(plane)")
        self.declare_answer("~means_of_transport(plane)")
        self.declare_answer("plane")
        self.declare_answer("~plane")
        self.add_answers_that_are_compatible_due_to_multiple_instance_predicate()
        self.declare_answer("implies(dept_city(paris), price(123.0))")
        self.declare_answer("implies(dest_city(paris), dest_city_type(resort))")

    def add_answers_that_are_compatible_due_to_multiple_instance_predicate(self):
        self.declare_answer("passenger_type_to_add(adult)")
        self.declare_answer("passenger_type_to_add(child)")

    def test_valid_combines(self):
        self.semantics_test_valid_combines()

    def test_invalid_combines(self):
        self.semantics_test_invalid_combines()

    def test_relevance(self):
        self.semantics_test_relevance()

    def test_irrelevance(self):
        self.semantics_test_irrelevance()

    def test_resolves(self):
        self.semantics_test_resolves()

    def test_non_resolves(self):
        self.semantics_test_non_resolves()

    def test_valid_incompatibles(self):
        self.semantics_test_valid_incompatibles()

    def test_invalid_incompatibles(self):
        self.semantics_test_invalid_incompatibles()


class TestSemanticLogicWithPropositions(unittest.TestCase):
    def setUp(self):
        self._ddd_component_manager = DDDComponentManager()
        self._create_ontology()
        self.ontology = self._ddd_component_manager.get_ontology(self.ontology_name)
        self._create_semantic_objects()
        self._language_code = "mocked_language"

        self.positive_prop_set = PropositionSet([self.proposition_dest_city_london, self.proposition_dest_city_paris])

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        self._city_sort = CustomSort(self.ontology_name, "city", dynamic=True)
        sorts = {
            self._city_sort,
            CustomSort(self.ontology_name, "city_type"),
            CustomSort(self.ontology_name, "ticket_type"),
            CustomSort(self.ontology_name, "passenger_type"),
        }
        predicates = {
            self._create_predicate("dest_city", self._city_sort),
            self._create_predicate("dept_city", self._city_sort),
        }
        individuals = {"paris": self._city_sort, "london": self._city_sort}
        actions = {"top", "buy"}
        ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)
        self._ddd_component_manager.add_ontology(ontology)

        self.empty_ontology = Ontology("empty_ontology", {}, {}, {}, set([]))
        self._ddd_component_manager.add_ontology(self.empty_ontology)

    def _create_predicate(self, *args, **kwargs):
        return Predicate(self.ontology_name, *args, **kwargs)

    def _create_semantic_objects(self):
        self.sort_city = self.ontology.get_sort("city")
        self.sort_city_type = self.ontology.get_sort("city_type")

        self.predicate_dest_city = self.ontology.get_predicate("dest_city")
        self.predicate_dept_city = self.ontology.get_predicate("dept_city")

        self.individual_paris = self.ontology.create_individual("paris")
        self.individual_london = self.ontology.create_individual("london")
        self.individual_not_paris = self.ontology.create_negative_individual("paris")

        self.proposition_dest_city_paris = \
            PredicateProposition(self.predicate_dest_city,
                                 self.individual_paris)
        self.proposition_dest_city_london = \
            PredicateProposition(self.predicate_dest_city,
                                 self.individual_london)
        self.proposition_dept_city_paris = \
            PredicateProposition(self.predicate_dept_city,
                                 self.individual_paris)
        self.proposition_dept_city_london = \
            PredicateProposition(self.predicate_dept_city,
                                 self.individual_london)
        self.proposition_not_dest_city_paris = \
            PredicateProposition(self.predicate_dest_city,
                                 self.individual_paris,
                                 Polarity.NEG)

    def test_combine_with_proposition_in_set_returns_proposition(self):
        self.assertEqual(
            self.proposition_dest_city_paris,
            SemanticLogic(self._ddd_component_manager
                          ).combine(self.positive_prop_set, self.proposition_dest_city_paris)
        )

    def test_combine_with_individual_in_set_returns_proposition(self):
        self.assertEqual(
            self.proposition_dest_city_paris,
            SemanticLogic(self._ddd_component_manager).combine(self.positive_prop_set, self.individual_paris)
        )

    def test_combine_with_negated_proposition_in_set_returns_negated_proposition(self):
        self.assertEqual(
            self.proposition_not_dest_city_paris,
            SemanticLogic(self._ddd_component_manager
                          ).combine(self.positive_prop_set, self.proposition_not_dest_city_paris)
        )

    def test_combine_with_negated_individual_in_set_returns_negated_proposition(self):
        self.assertEqual(
            self.proposition_not_dest_city_paris,
            SemanticLogic(self._ddd_component_manager).combine(self.positive_prop_set, self.individual_not_paris)
        )



# -*- coding: utf-8 -*-

import pytest
from unittest.mock import Mock

from tala.model.common import Modality
from tala.model.speaker import Speaker
from tala.model.move import Move, ICMMove, ReportMove, ICMMoveWithSemanticContent, AskMove
from tala.model.proposition import ServiceResultProposition, PredicateProposition
from tala.testing.lib_test_case import LibTestCase
from tala.model.service_action_outcome import SuccessfulServiceAction, FailedServiceAction
from tala.testing.move_factory import MoveFactoryWithPredefinedBoilerplate


class SemanticExpressionTestMixin(object):
    def given_move(self, move):
        self._move = move

    def when_generating_semantic_expression(self):
        self._actual_result = self._move.as_semantic_expression()

    def given_set_realization_data_was_called(self, **kwargs):
        self._move.set_realization_data(**kwargs)

    def when_get_semantic_expression_without_realization_data(self):
        self._actual_result = self._move.semantic_expression_without_realization_data()


class ReportMoveTests(LibTestCase, SemanticExpressionTestMixin):
    def setUp(self):
        self.setUpLibTestCase()
        self.move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name, ddd_name="mockup_ddd")
        self.service_action = "BuyTrip"
        self.parameters = [self.proposition_dest_city_paris, self.proposition_dept_city_london]
        self.success_result = ServiceResultProposition(
            self.ontology_name, self.service_action, self.parameters, SuccessfulServiceAction()
        )
        self.success_move = self.move_factory.create_report_move(self.success_result)
        self.failure_move = self.move_factory.create_report_move(
            ServiceResultProposition(
                self.ontology_name, self.service_action, self.parameters, FailedServiceAction("no_itinerary_found")
            )
        )

    def test_getters(self):
        self.assertEqual(Move.REPORT, self.success_move.type_)
        self.assertEqual(self.success_result, self.success_move.get_content())

    def test_str_success_move(self):
        self.assertEqual(
            "report(ServiceResultProposition("
            "BuyTrip, [dest_city(paris), dept_city(london)], SuccessfulServiceAction()))", str(self.success_move)
        )

    def test_str_failure_move(self):
        self.assertEqual(
            "report(ServiceResultProposition("
            "BuyTrip, [dest_city(paris), dept_city(london)], FailedServiceAction(no_itinerary_found)))",
            str(self.failure_move)
        )

    def test_str_with_ddd_name(self):
        self.success_move.set_ddd_name("mockup_ddd")
        self.assertEqual(
            "report(ServiceResultProposition("
            "BuyTrip, [dest_city(paris), dept_city(london)], SuccessfulServiceAction()), ddd_name='mockup_ddd')",
            str(self.success_move)
        )

    def test_equality(self):
        move = self.move_factory.create_report_move(
            ServiceResultProposition(
                self.ontology_name, self.service_action, self.parameters, SuccessfulServiceAction()
            )
        )
        identical_move = self.move_factory.create_report_move(
            ServiceResultProposition(
                self.ontology_name, self.service_action, self.parameters, SuccessfulServiceAction()
            )
        )
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(move, identical_move)

    def test_unicode_with_swedish_character(self):
        parameters = [
            PredicateProposition(
                self.ontology.get_predicate("comment_message"), self.ontology.create_individual("balkong önskas")
            )
        ]
        move = self.move_factory.create_report_move(
            ServiceResultProposition(self.ontology_name, "RegisterComment", parameters, SuccessfulServiceAction())
        )
        self.assertEqual(
            'report(ServiceResultProposition('
            'RegisterComment, [comment_message("balkong önskas")], SuccessfulServiceAction()))', str(move)
        )

    def test_as_semantic_expression(self):
        self.given_move(self.success_move)
        self.when_generating_semantic_expression()
        self.then_result_is(
            "report("
            "ServiceResultProposition(BuyTrip, [dest_city(paris), dept_city(london)], SuccessfulServiceAction())"
            ")"
        )

    def test_semantic_expression_without_realization_data(self):
        self.given_move(self.success_move)
        self.given_set_realization_data_was_called(speaker=Speaker.SYS, ddd_name="mock_ddd")
        self.when_get_semantic_expression_without_realization_data()
        self.then_result_is(
            "report(ServiceResultProposition(BuyTrip, [dest_city(paris), dept_city(london)], SuccessfulServiceAction()))"
        )


class PrereportMoveTests(LibTestCase, SemanticExpressionTestMixin):
    def setUp(self):
        self.setUpLibTestCase()
        self.move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name, ddd_name="mockup_ddd")
        self.service_action = "BuyTrip"
        self.arguments = [self.proposition_dest_city_paris, self.proposition_dept_city_london]
        self.move = self.move_factory.create_prereport_move(self.service_action, self.arguments)

    def test_getters(self):
        self.assertEqual(Move.PREREPORT, self.move.type_)
        self.assertEqual(self.service_action, self.move.get_service_action())
        self.assertEqual(self.arguments, self.move.get_arguments())

    def test_to_string(self):
        self.assertEqual("prereport(BuyTrip, [dest_city(paris), dept_city(london)])", str(self.move))

    def test_equality(self):
        move = self.move_factory.create_prereport_move(self.service_action, self.arguments)
        identical_move = self.move_factory.create_prereport_move(self.service_action, self.arguments)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(move, identical_move)

    def test_inequality_due_to_arguments(self):
        move1 = self.move_factory.create_prereport_move(self.service_action, [])
        move2 = self.move_factory.create_prereport_move(self.service_action, [self.proposition_dest_city_paris])
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(move1, move2)

    def test_inequality_due_to_service_action(self):
        move1 = self.move_factory.create_prereport_move("action1", [])
        move2 = self.move_factory.create_prereport_move("action2", [])
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(move1, move2)

    def test_as_semantic_expression(self):
        self.given_move(self.move)
        self.when_generating_semantic_expression()
        self.then_result_is("prereport(BuyTrip, [dest_city(paris), dept_city(london)])")

    def test_semantic_expression_without_realization_data(self):
        self.given_move(self.move_factory.create_prereport_move(self.service_action, self.arguments))
        self.given_set_realization_data_was_called(speaker=Speaker.SYS, ddd_name="mock_ddd")
        self.when_get_semantic_expression_without_realization_data()
        self.then_result_is("prereport(BuyTrip, [dest_city(paris), dept_city(london)])")


class MoveTests(LibTestCase, SemanticExpressionTestMixin):
    def setUp(self):
        self.setUpLibTestCase()
        self.move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name, ddd_name="mockup_ddd")
        self.icm_acc_neg = self.move_factory.createIcmMove(ICMMove.ACC, speaker=Speaker.SYS, polarity=ICMMove.NEG)
        self.icm_acc_neg_issue = self.move_factory.createIcmMove(
            ICMMove.ACC, speaker=Speaker.SYS, polarity=ICMMove.NEG, content="issue"
        )
        self.icm_per_neg = self.move_factory.createIcmMove(ICMMove.PER, speaker=Speaker.SYS, polarity=ICMMove.NEG)
        self.icm_per_pos = self.move_factory.createIcmMove(ICMMove.PER, speaker=Speaker.SYS, polarity=ICMMove.POS)
        self.icm_acc_pos = self.move_factory.createIcmMove(ICMMove.ACC, speaker=Speaker.SYS, polarity=ICMMove.POS)
        self.report_move = self.move_factory.create_report_move(
            ServiceResultProposition(self.ontology_name, "service_action", [], SuccessfulServiceAction())
        )
        self.answer_move = self.move_factory.createAnswerMove(speaker=Speaker.SYS, answer=self.individual_paris)
        self.request_move = self.move_factory.createRequestMove(action=self.buy_action)
        self.ask_move = self.move_factory.create_ask_move(speaker=Speaker.SYS, question=self.price_question)

    def test_as_semantic_expression_for_answer(self):
        self.given_move(self.answer_move)
        self.when_generating_semantic_expression()
        self.then_result_is("answer(paris)")

    def test_as_semantic_expression_for_request(self):
        self.given_move(self.request_move)
        self.when_generating_semantic_expression()
        self.then_result_is("request(buy)")

    def test_as_semantic_expression_for_ask(self):
        self.given_move(self.ask_move)
        self.when_generating_semantic_expression()
        self.then_result_is("ask(?X.price(X))")

    def test_as_semantic_expression_for_icm_acc_neg(self):
        self.given_move(self.icm_acc_neg)
        self.when_generating_semantic_expression()
        self.then_result_is("icm:acc*neg")

    def test_as_semantic_expression_for_icm_acc_neg_issue(self):
        self.given_move(self.icm_acc_neg_issue)
        self.when_generating_semantic_expression()
        self.then_result_is("icm:acc*neg:issue")

    def test_as_semantic_expression_for_icm_with_propositional_content(self):
        self.given_move(
            self.move_factory.createIcmMove(
                ICMMove.SEM, speaker=Speaker.SYS, polarity=ICMMove.NEG, content=self.proposition_dest_city_paris
            )
        )
        self.when_generating_semantic_expression()
        self.then_result_is("icm:sem*neg:dest_city(paris)")

    def test_as_semantic_expression_for_icm_per_neg(self):
        self.given_move(self.icm_per_neg)
        self.when_generating_semantic_expression()
        self.then_result_is("icm:per*neg")

    def test_as_semantic_expression_for_icm_per_pos(self):
        self.given_move(self.icm_per_pos)
        self.when_generating_semantic_expression()
        self.then_result_is("icm:per*pos")

    def test_as_semantic_expression_for_icm_acc_pos(self):
        self.given_move(self.icm_acc_pos)
        self.when_generating_semantic_expression()
        self.then_result_is("icm:acc*pos")

    def testShortAnswerunicode(self):
        move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name)
        answer = move_factory.createAnswerMove(self.individual_paris)
        self.assertEqual("Move(answer(paris))", str(answer))

    def test_short_answer_is_answer_move(self):
        move_factory = MoveFactoryWithPredefinedBoilerplate(
            self.ontology_name, speaker="USR", understanding_confidence=1.0, ddd_name=""
        )
        answer = move_factory.createAnswerMove(self.individual_paris)
        self.assertEqual(Move.ANSWER, answer.type_)

    def test_is_question_raising_true_for_ask_move(self):
        question_string = "?X.dest_city(X)"
        move_factory = MoveFactoryWithPredefinedBoilerplate(
            self.ontology_name, speaker="USR", understanding_confidence=1.0, ddd_name=""
        )
        ask_move = move_factory.create_ask_move(question_string)
        self.assertTrue(ask_move.is_question_raising())

    def test_is_question_raising_false_for_answer_move(self):
        proposition_string = "dest_city(paris)"
        move_factory = MoveFactoryWithPredefinedBoilerplate(
            self.ontology_name, speaker="USR", understanding_confidence=1.0, ddd_name=""
        )
        answer_move = move_factory.createAnswerMove(proposition_string)
        self.assertFalse(answer_move.is_question_raising())

    def test_is_question_raising_true_for_contentful_icm_und_neg(self):
        propositional_content = self.proposition_dest_city_paris
        contentful_icm_und_neg = ICMMoveWithSemanticContent(ICMMove.UND, propositional_content, polarity=ICMMove.NEG)
        self.assertTrue(contentful_icm_und_neg.is_question_raising())

    def test_is_question_raising_false_for_contentless_icm_und_neg(self):
        contentless_icm_und_neg = ICMMove(ICMMove.UND, polarity=ICMMove.NEG)
        self.assertFalse(contentless_icm_und_neg.is_question_raising())

    def test_is_question_raising_true_for_icm_und_int(self):
        propositional_content = self.proposition_dest_city_paris
        icm_und_int = ICMMoveWithSemanticContent(ICMMove.UND, propositional_content, polarity=ICMMove.INT)
        self.assertTrue(icm_und_int.is_question_raising())

    def test_is_question_raising_true_for_icm_und_pos_for_positive_content(self):
        positive_content = self.proposition_dest_city_paris
        icm = ICMMoveWithSemanticContent(ICMMove.UND, positive_content, polarity=ICMMove.POS)
        self.assertTrue(icm.is_question_raising())

    def test_is_question_raising_false_for_icm_und_pos_for_negated_content(self):
        negation = self.proposition_not_dest_city_paris
        icm = ICMMoveWithSemanticContent(ICMMove.UND, negation, polarity=ICMMove.POS)
        self.assertFalse(icm.is_question_raising())

    def test_is_question_raising_false_for_icm_acc_neg(self):
        icm_acc_neg = ICMMove(ICMMove.ACC, polarity=ICMMove.NEG)
        self.assertFalse(icm_acc_neg.is_question_raising())

    def test_create_greet_move(self):
        move = self.move_factory.createMove(Move.GREET, speaker=Speaker.SYS)
        self.assertEqual(Move.GREET, move.type_)

    def test_greet_move_to_string(self):
        move = self.move_factory.createMove(Move.GREET, speaker=Speaker.SYS)
        self.assertEqual(
            "Move(greet, ddd_name='mockup_ddd', speaker=SYS, understanding_confidence=1.0, "
            "weighted_understanding_confidence=1.0, perception_confidence=1.0, modality=speech)", str(move)
        )

    def test_greet_move_to_semantic_expression(self):
        move = self.move_factory.createMove(Move.GREET, speaker=Speaker.SYS)
        self.assertEqual("greet", move.as_semantic_expression())

    def test_string_representation_with_score_and_speaker(self):
        move = self.move_factory.createMove(
            Move.GREET, understanding_confidence=1.0, speaker=Speaker.SYS, perception_confidence=0.9
        )
        move_string = "Move(greet, ddd_name='mockup_ddd', speaker=SYS, understanding_confidence=1.0, " "weighted_understanding_confidence=1.0, perception_confidence=0.9, modality=speech)"
        self.assertEqual(move_string, str(move))

    def test_str_with_background(self):
        move = self.move_factory.create_ask_move(speaker=Speaker.SYS, question=self.dept_city_question)
        move.set_background([self.proposition_dest_city_paris])
        self.assertEqual(
            "Move(ask(?X.dept_city(X), [dest_city(paris)]), ddd_name='mockup_ddd', speaker=SYS, "
            "understanding_confidence=1.0, weighted_understanding_confidence=1.0, "
            "perception_confidence=1.0, modality=speech)", str(move)
        )

    def test_str_with_utterance(self):
        move = self.move_factory.createMove(
            Move.GREET, understanding_confidence=0.5, speaker=Speaker.USR, utterance="hello"
        )
        move_string = "Move(greet, ddd_name='mockup_ddd', speaker=USR, understanding_confidence=0.5, " "weighted_understanding_confidence=0.5, perception_confidence=1.0, modality=speech, " "utterance='hello')"
        self.assertEqual(move_string, str(move))

    def test_move_equality_basic(self):
        move1 = self.move_factory.createMove(
            Move.GREET, speaker=Speaker.USR, perception_confidence=0.8, understanding_confidence=0.6
        )
        move2 = self.move_factory.createMove(
            Move.GREET, speaker=Speaker.USR, perception_confidence=0.8, understanding_confidence=0.6
        )
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(move1, move2)

    def test_move_inequality_basic(self):
        move1 = self.move_factory.createMove(Move.QUIT, speaker=Speaker.SYS)
        move2 = self.move_factory.createMove(Move.GREET, speaker=Speaker.SYS)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(move1, move2)

    def test_speaker_inequality(self):
        user_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=1.0)
        sys_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.SYS, understanding_confidence=1.0)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(user_move, sys_move)

    def test_understanding_confidence_inequality(self):
        hi_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=0.9)
        lo_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=0.8)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(lo_move, hi_move)

    def test_weighted_understanding_confidence_inequality(self):
        hi_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=0.5)
        hi_move.weighted_understanding_confidence = 0.9
        lo_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=0.5)
        lo_move.weighted_understanding_confidence = 0.8
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(lo_move, hi_move)

    def test_perception_confidence_inequality(self):
        hi_move = self.move_factory.createMove(
            Move.GREET, speaker=Speaker.USR, understanding_confidence=0.5, perception_confidence=0.5
        )
        lo_move = self.move_factory.createMove(
            Move.GREET, speaker=Speaker.USR, understanding_confidence=0.5, perception_confidence=0.6
        )
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(lo_move, hi_move)

    def test_utterance_inequality(self):
        move1 = self.move_factory.createMove(
            Move.GREET, speaker=Speaker.USR, understanding_confidence=1.0, utterance="hello"
        )
        move2 = self.move_factory.createMove(
            Move.GREET, speaker=Speaker.USR, understanding_confidence=1.0, utterance="hi"
        )
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(move1, move2)

    def test_move_content_equality_tolerates_score_diff(self):
        hi_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=0.9)
        lo_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=0.8)
        self.assertTrue(hi_move.move_content_equals(lo_move))
        self.assertTrue(lo_move.move_content_equals(hi_move))

    def test_move_content_equality_tolerates_speaker_diff(self):
        usr_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=1.0)
        sys_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.SYS, understanding_confidence=1.0)
        self.assertTrue(sys_move.move_content_equals(usr_move))
        self.assertTrue(usr_move.move_content_equals(sys_move))

    def test_move_content_inequality_type_differs(self):
        quit_move = self.move_factory.createMove(Move.QUIT, speaker=Speaker.SYS)
        greet_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.SYS)
        self.assertFalse(quit_move.move_content_equals(greet_move))
        self.assertFalse(greet_move.move_content_equals(quit_move))

    def test_content_inequality_content_differs(self):
        paris_answer_move = self.move_factory.createAnswerMove(speaker=Speaker.SYS, answer=self.individual_paris)
        london_answer_move = self.move_factory.createAnswerMove(speaker=Speaker.SYS, answer=self.individual_london)
        self.assertFalse(paris_answer_move.move_content_equals(london_answer_move))

    def test_content_inequality(self):
        move1 = self.move_factory.createAnswerMove(speaker=Speaker.SYS, answer=self.individual_paris)
        move2 = self.move_factory.createAnswerMove(speaker=Speaker.SYS, answer=self.individual_london)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(move1, move2)

    def test_move_inequality_between_move_classes(self):
        icm = self.move_factory.createIcmMove(ICMMove.ACC, content=ICMMove.POS)
        report = ReportMove(ServiceResultProposition(self.ontology_name, "action", [], SuccessfulServiceAction()))
        self.assertFalse(report.move_content_equals(icm))

    def test_is_negative_perception_icm(self):
        self.assertTrue(self.icm_per_neg.is_negative_perception_icm())
        self.assertFalse(self.icm_per_pos.is_negative_perception_icm())

    def test_is_positive_acceptance_icm(self):
        self.assertTrue(self.icm_acc_pos.is_positive_acceptance_icm())
        self.assertFalse(self.icm_per_pos.is_positive_acceptance_icm())

    def test_is_negative_acceptance_issue_icm(self):
        self.assertTrue(self.icm_acc_neg_issue.is_negative_acceptance_issue_icm())
        self.assertFalse(self.icm_per_neg.is_negative_acceptance_issue_icm())
        self.assertFalse(self.icm_acc_neg.is_negative_acceptance_issue_icm())

    def test_create_prereport_move(self):
        service_action = "BuyTrip"
        dest_city = self.proposition_dest_city_paris
        dept_city = self.proposition_dept_city_london
        parameters = [dept_city, dest_city]
        self.move_factory.create_prereport_move(service_action, parameters)

    def test_is_turn_yielding(self):
        turn_yielding_moves = [self.report_move, self.answer_move]
        non_turn_yielding_moves = [self.icm_acc_pos, self.icm_acc_neg, self.request_move, self.ask_move]
        for move in turn_yielding_moves:
            self.assertTrue(move.is_turn_yielding(), f"{move} should be turn yielding")
        for move in non_turn_yielding_moves:
            self.assertFalse(move.is_turn_yielding(), f"{move} should not be turn yielding")

    def test_upscore(self):
        move = self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=0.5)
        move.uprank(0.2)
        self.assertEqual(0.5 * 1.2, move.weighted_understanding_confidence)

    def test_set_and_get_weighted_understanding_confidence_without_setting_realization_data(self):
        self.given_move(self.move_factory.createMove(Move.GREET, speaker=Speaker.USR, understanding_confidence=0.5))
        self.given_weighted_understanding_confidence_was_set(0.6)
        self.when_get_weighted_understanding_confidence()
        self.then_result_is(0.6)

    def given_weighted_understanding_confidence_was_set(self, score):
        self._move.weighted_understanding_confidence = score

    def when_get_weighted_understanding_confidence(self):
        self._actual_result = self._move.weighted_understanding_confidence

    def test_set_and_get_weighted_understanding_confidence_after_setting_realization_data(self):
        self.given_move(Move(Move.GREET))
        self.given_set_realization_data_was_called(
            speaker=Speaker.USR, ddd_name="mock_ddd", understanding_confidence=0.5
        )
        self.given_weighted_understanding_confidence_was_set(0.6)
        self.when_get_weighted_understanding_confidence()
        self.then_result_is(0.6)

    def test_semantic_expression_without_realization_data_for_move_without_content(self):
        self.given_move(Move(Move.GREET))
        self.given_set_realization_data_was_called(
            speaker=Speaker.USR, ddd_name="mock_ddd", understanding_confidence=0.5
        )
        self.when_get_semantic_expression_without_realization_data()
        self.then_result_is("Move(greet)")

    def test_semantic_expression_without_realization_data_for_move_with_content(self):
        self.given_move(AskMove(self.price_question))
        self.given_set_realization_data_was_called(speaker=Speaker.SYS, ddd_name="mock_ddd")
        self.when_get_semantic_expression_without_realization_data()
        self.then_result_is("Move(ask(?X.price(X)))")


class ICMMoveTests(LibTestCase, SemanticExpressionTestMixin):
    def setUp(self):
        self.setUpLibTestCase()
        self.move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name)

    def test_create_reraise_icm_move(self):
        move = self.move_factory.createIcmMove(ICMMove.RERAISE, content=self.ontology.create_action("top"))
        self.assertTrue(move.type_ == ICMMove.RERAISE)

    def test_reraise_to_string(self):
        move = self.move_factory.createIcmMove(ICMMove.RERAISE, content=self.ontology.create_action("top"))
        self.assertEqual("ICMMove(icm:reraise:top)", str(move))

    def test_loadplan_to_string(self):
        move = self.move_factory.createIcmMove(ICMMove.LOADPLAN)
        self.assertEqual("ICMMove(icm:loadplan)", str(move))

    def test_string_representation_with_score_and_speaker(self):
        move = self.move_factory.createIcmMove(
            ICMMove.ACC,
            polarity=ICMMove.POS,
            understanding_confidence=1.0,
            speaker=Speaker.SYS,
            perception_confidence=0.9
        )
        self.assertEqual(
            "ICMMove(icm:acc*pos, speaker=SYS, understanding_confidence=1.0, perception_confidence=0.9)", str(move)
        )

    def test_create_und_int(self):
        icm = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.INT, speaker=Speaker.SYS)
        self.assertEqual(ICMMove.UND, icm.type_)
        self.assertEqual(ICMMove.INT, icm.get_polarity())

    def test_inequality_due_to_polarity_difference(self):
        icm_int = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.INT)
        icm_pos = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.POS)
        icm_neg = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.NEG)
        self.assertNotEqual(icm_pos, icm_int)
        self.assertNotEqual(icm_pos, icm_neg)
        self.assertNotEqual(icm_int, icm_neg)

    def test_inequality_due_to_type_difference(self):
        icm = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.POS)
        non_identical_icm = self.move_factory.createIcmMove(ICMMove.ACC, polarity=ICMMove.POS)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(icm, non_identical_icm)

    def test_inequality_due_to_content_difference(self):
        icm = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.POS, content="dummy_content")
        non_identical_icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, content="different_dummy_content"
        )
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(icm, non_identical_icm)

    def test_inequality_due_to_content_speaker_difference(self):
        icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, content=Mock(), content_speaker=Speaker.USR
        )
        non_identical_icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, content=Mock(), content_speaker=Speaker.SYS
        )
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(icm, non_identical_icm)

    def test_inequality_due_to_speaker_difference(self):
        icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, understanding_confidence=1.0, speaker=Speaker.SYS
        )
        non_identical_icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.NEG, understanding_confidence=1.0, speaker=Speaker.SYS
        )
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(icm, non_identical_icm)

    def test_inequality_due_to_score_difference(self):
        icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, speaker=Speaker.USR, understanding_confidence=0.9, ddd_name="mockup_ddd"
        )
        non_identical_icm = self.move_factory.createIcmMove(
            ICMMove.UND,
            polarity=ICMMove.POS,
            speaker=Speaker.USR,
            understanding_confidence=0.77,
            ddd_name="mockup_ddd"
        )
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(icm, non_identical_icm)

    def test_content_inequality_due_to_polarity_difference(self):
        icm_int = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.INT)
        icm_pos = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.POS)
        self.assertFalse(icm_pos.move_content_equals(icm_int))

    def test_content_inequality_due_to_type_difference(self):
        icm = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.POS)
        non_identical_icm = self.move_factory.createIcmMove(ICMMove.ACC, polarity=ICMMove.POS)
        self.assertFalse(icm.move_content_equals(non_identical_icm))

    def test_content_inequality_due_to_content_difference(self):
        icm = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.POS, content="dummy_content")
        non_identical_icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, content="different_dummy_content"
        )
        self.assertFalse(icm.move_content_equals(non_identical_icm))

    def test_content_inequality_due_to_content_speaker_difference(self):
        icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, content=Mock(), content_speaker=Speaker.USR
        )
        non_identical_icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, content=Mock(), content_speaker=Speaker.SYS
        )
        self.assertFalse(icm.move_content_equals(non_identical_icm))

    def test_content_equality_tolerates_speaker_difference(self):
        icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, understanding_confidence=1.0, speaker=Speaker.USR, ddd_name="mockup_ddd"
        )
        non_identical_icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, understanding_confidence=1.0, speaker=Speaker.SYS, ddd_name="mockup_ddd"
        )
        self.assertTrue(icm.move_content_equals(non_identical_icm))

    def test_content_equality_tolerates_score_difference(self):
        icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, speaker=Speaker.USR, understanding_confidence=0.9, ddd_name="mockup_ddd"
        )
        non_identical_icm = self.move_factory.createIcmMove(
            ICMMove.UND,
            polarity=ICMMove.POS,
            speaker=Speaker.USR,
            understanding_confidence=0.77,
            ddd_name="mockup_ddd"
        )
        self.assertTrue(icm.move_content_equals(non_identical_icm))

    def test_und_neg_to_string(self):
        icm = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.NEG)
        expected_string = "ICMMove(icm:und*neg)"
        self.assertEqual(expected_string, str(icm))

    def test_und_int_to_string(self):
        icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.INT, content_speaker=Speaker.USR, content=self.proposition_dest_city_paris
        )
        expected_string = "ICMMove(icm:und*int:USR*dest_city(paris))"
        self.assertEqual(expected_string, str(icm))

    def test_create_und_pos(self):
        icm = self.move_factory.createIcmMove(ICMMove.UND, polarity=ICMMove.POS, speaker=Speaker.SYS)
        self.assertEqual(ICMMove.UND, icm.type_)
        self.assertEqual(ICMMove.POS, icm.get_polarity())

    def test_und_pos_to_string(self):
        icm = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.POS, content_speaker=Speaker.USR, content=self.proposition_dest_city_paris
        )
        expected_string = "ICMMove(icm:und*pos:USR*dest_city(paris))"
        self.assertEqual(expected_string, str(icm))

    def test_per_neg_to_string(self):
        icm = self.move_factory.createIcmMove(ICMMove.PER, polarity=ICMMove.NEG)
        self.assertEqual("ICMMove(icm:per*neg)", str(icm))

    def test_per_pos_to_string_with_content_as_double_quoted_string(self):
        icm = self.move_factory.createIcmMove(ICMMove.PER, polarity=ICMMove.POS, content='a string')
        self.assertEqual('ICMMove(icm:per*pos:"a string")', str(icm))

    def test_acc_pos_to_string(self):
        icm = self.move_factory.createIcmMove(ICMMove.ACC, polarity=ICMMove.POS)
        self.assertEqual("ICMMove(icm:acc*pos)", str(icm))

    def test_sem_neg_to_string(self):
        icm = self.move_factory.createIcmMove(ICMMove.SEM, polarity=ICMMove.NEG)
        expected_string = "ICMMove(icm:sem*neg)"
        self.assertEqual(expected_string, str(icm))

    def test_acc_neg_issue_to_string(self):
        icm = self.move_factory.createIcmMove(ICMMove.ACC, polarity=ICMMove.NEG, content="issue")
        self.assertEqual("ICMMove(icm:acc*neg:issue)", str(icm))

    def test_icm_move_getters(self):
        move = self.move_factory.createIcmMove(
            ICMMove.UND, polarity=ICMMove.INT, content=self.price_question, content_speaker=Speaker.USR
        )
        self.assertTrue(move.is_icm())
        self.assertEqual(ICMMove.INT, move.get_polarity())
        self.assertEqual(self.price_question, move.get_content())
        self.assertEqual(Speaker.USR, move.get_content_speaker())

    def test_semantic_expression_without_realization_data(self):
        self.given_move(self.move_factory.createIcmMove(ICMMove.SEM, polarity=ICMMove.NEG))
        self.given_set_realization_data_was_called(speaker=Speaker.SYS, ddd_name="mock_ddd")
        self.when_get_semantic_expression_without_realization_data()
        self.then_result_is("ICMMove(icm:sem*neg)")

    def test_ddd_getter_and_setter(self):
        move = self.move_factory.createMove(Move.GREET)
        name = "mockup_ddd"
        move.set_ddd_name(name)
        self.assertEqual(name, move.get_ddd_name())


class MoveRealizationTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name)

    def test_set_realization_data_for_haptic_input(self):
        move = self.move_factory.createMove(Move.GREET)
        understanding_confidence = 0.5
        speaker = Speaker.USR
        modality = Modality.HAPTIC
        move.set_realization_data(
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            modality=modality,
            ddd_name="mockup_ddd"
        )
        self.assertEqual(understanding_confidence, move.understanding_confidence)
        self.assertEqual(speaker, move.get_speaker())
        self.assertEqual(modality, move.get_modality())

    def test_set_realization_data_for_spoken_input(self):
        move = self.move_factory.createMove(Move.GREET)
        understanding_confidence = 0.5
        speaker = Speaker.USR
        modality = Modality.SPEECH
        utterance = "hello"
        move.set_realization_data(
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            modality=modality,
            utterance=utterance,
            ddd_name="mockup_ddd"
        )
        self.assertEqual(understanding_confidence, move.understanding_confidence)
        self.assertEqual(speaker, move.get_speaker())
        self.assertEqual(modality, move.get_modality())
        self.assertEqual(utterance, move.get_utterance())

    def test_get_score_returns_one_for_move_with_speaker_sys(self):
        greet_move = self.move_factory.createMove(Move.GREET, speaker=Speaker.SYS)
        icm_move = self.move_factory.createIcmMove(ICMMove.ACC, content=ICMMove.POS, speaker=Speaker.SYS)
        for sys_move in [greet_move, icm_move]:
            self.assertEqual(1.0, sys_move.understanding_confidence)

    def test_move_speech_modality_is_default(self):
        quit_move = self.move_factory.createMove(Move.QUIT, speaker=Speaker.SYS)
        icm_move = self.move_factory.createIcmMove(ICMMove.ACC, content=ICMMove.POS, speaker=Speaker.SYS)
        for move in [quit_move, icm_move]:
            self.assertEqual(Modality.SPEECH, move.get_modality())

    def test_move_haptic_modality_is_available(self):
        self.move_factory.createMove(Move.QUIT, speaker=Speaker.SYS, modality=Modality.HAPTIC)

    def test_move_text_modality_is_available(self):
        self.move_factory.createMove(Move.QUIT, speaker=Speaker.SYS, modality=Modality.TEXT)


class MoveTestsFromTdmLib(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def testPropositionalAnswerMoveEquality(self):
        move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name)
        destination_answer = self.proposition_dest_city_paris
        answerMove = move_factory.createAnswerMove(destination_answer)
        duplicateMove = move_factory.createAnswerMove(destination_answer)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(answerMove, duplicateMove)

    def testYesNoAskMoveunicode(self):
        question = self.ontology.create_yes_no_question("dest_city", "paris")
        move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name)
        move = move_factory.create_ask_move(question)
        moveAsString = str(move)
        self.assertEqual("Move(ask(?dest_city(paris)))", moveAsString)

    def test_wh_ask_move_to_string(self):
        move_factory = MoveFactoryWithPredefinedBoilerplate(self.ontology_name)
        move = move_factory.create_ask_move(self.dest_city_question)
        moveAsString = str(move)
        self.assertEqual("Move(ask(?X.dest_city(X)))", moveAsString)


class TestICMMoves(object):
    def setup_method(self):
        self._move = None
        self._actual_result = None

    @pytest.mark.parametrize("move,expected_result", [
        (ICMMoveWithSemanticContent(ICMMove.UND, content="content", polarity=ICMMove.INT), True),
        (ICMMoveWithSemanticContent(ICMMove.UND, content="content", polarity=ICMMove.POS), True),
        (ICMMoveWithSemanticContent(ICMMove.UND, content="content", polarity=ICMMove.NEG), False),
        (ICMMoveWithSemanticContent(ICMMove.SEM, content="content", polarity=ICMMove.POS), False),
        (ICMMoveWithSemanticContent(ICMMove.PER, content="content", polarity=ICMMove.POS), False),
        (ICMMove(ICMMove.UND, polarity=ICMMove.INT), False),
        (ICMMove(ICMMove.UND, polarity=ICMMove.POS), False)
    ])  # yapf: disable
    def test_is_grounding_move(self, move, expected_result):
        self.given_move(move)
        self.when_calling_is_grounding_move()
        self.then_result_is(expected_result)

    def given_move(self, move):
        self._move = move

    def when_calling_is_grounding_move(self):
        self._actual_result = self._move.is_grounding_proposition()

    def then_result_is(self, expected_result):
        assert self._actual_result == expected_result



from tala.utils.as_json import AsJSONMixin
from tala.utils.equality import EqualityMixin


class DDD(AsJSONMixin, EqualityMixin):
    def __init__(self, name, ontology, domain, service_interface):
        super(DDD, self).__init__()
        self._name = name
        self._ontology = ontology
        self._domain = domain
        self._service_interface = service_interface

    @property
    def name(self):
        return self._name

    @property
    def ontology(self):
        return self._ontology

    @property
    def domain(self):
        return self._domain

    @property
    def service_interface(self):
        return self._service_interface

    def __repr__(self):
        return "%s%s" % (self.__class__.__name__, (self.name, self.ontology, self.domain, self.service_interface))

    def as_dict(self):
        return {
            "ddd_name": self._name,
            "ontology": self.ontology,
            "domain": self.domain,
            "service_interface": self.service_interface,
        }



import unittest

from mock import Mock

from tala.model.predicate import Predicate
from tala.model.sort import Sort
from tala.model.individual import Individual
from tala.model.proposition import Proposition, PredicateProposition
from tala.model.polarity import Polarity
from tala.ddd.ddd_component_manager import DDDComponentManager

from tala.http_formatting import HttpFormatter
from tala.ddd.ddd_specific_components import DDDSpecificComponents


class HttpFormattingTest(unittest.TestCase):
    def setUp(self):
        self._mock_ddd = self._create_mock_ddd_components()
        self._mock_ddd_manager = self._create_mock_ddd_component_manager()

    def _create_mock_ddd_components(self):
        mock_components = Mock(spec=DDDSpecificComponents)
        return mock_components

    def _create_mock_ddd_component_manager(self):
        mock_component_manager = Mock(spec=DDDComponentManager)
        mock_component_manager.get_ddd_specific_components_of_ontology_name.return_value = self._mock_ddd
        return mock_component_manager

    def test_fact_to_json_object_for_value_with_grammar_entry(self):
        self.given_created_http_formatter()
        self.when_call_fact_to_json_object(
            self.mock_predicate_proposition(
                predicate_name="mock_predicate",
                sort_name="mock_sort",
                value="mock_individual_value",
                value_as_json_object={"value": "mock_value"}
            ), {
                "session_id":
                "blah",
                "entities": [{
                    "sort": "mock_sort",
                    "natural_language_form": "mock_grammar_entry",
                    "name": "mock_individual_value"
                }]
            }
        )
        self.then_result_is({
            "sort": "mock_sort",
            "value": "mock_value",
            "grammar_entry": "mock_grammar_entry",
            'perception_confidence': None,
            'understanding_confidence': None,
            'weighted_confidence': None,
            'weighted_understanding_confidence': None
        })

    def given_created_http_formatter(self):
        self._http_formatter = HttpFormatter(self._mock_ddd_manager)

    def when_call_fact_to_json_object(self, fact, session):
        self._actual_result = self._http_formatter.fact_to_json_object(fact, session)

    def mock_predicate_proposition(
        self,
        predicate_name="mock_predicate",
        sort_name="mock_sort",
        value="mock_value",
        perception_confidence=None,
        understanding_confidence=None,
        weighted_confidence=None,
        weighted_understanding_confidence=None,
        value_as_json_object=None,
        polarity=Polarity.POS
    ):
        value_as_json_object = value_as_json_object or {"value": "mock_value"}
        sort = self._mock_sort(sort_name)
        predicate = self._mock_predicate(predicate_name, sort)
        individual = self._mock_individual(value, value_as_json_object, sort)
        mock_proposition = Mock(spec=PredicateProposition)
        mock_proposition.is_predicate_proposition.return_value = True
        mock_proposition.getPredicate.return_value = predicate
        mock_proposition.getArgument.return_value = individual
        mock_proposition.get_polarity.return_value = polarity
        mock_proposition.confidence_estimates.perception_confidence = perception_confidence
        mock_proposition.confidence_estimates.understanding_confidence = understanding_confidence
        mock_proposition.confidence_estimates.weighted_confidence = weighted_confidence
        mock_proposition.confidence_estimates.weighted_understanding_confidence = weighted_understanding_confidence
        return mock_proposition

    def _mock_predicate(self, name, sort):
        mock_predicate = Mock(spec=Predicate)
        mock_predicate.get_name.return_value = name
        mock_predicate.getSort.return_value = sort
        return mock_predicate

    def _mock_sort(self, name):
        mock_sort = Mock(spec=Sort)
        mock_sort.get_name.return_value = name
        return mock_sort

    def _mock_individual(self, value, value_as_json_object, sort):
        mock_individual = Mock(spec=Individual)
        mock_individual.getSort.return_value = sort
        mock_individual.getValue.return_value = value
        mock_individual.value_as_json_object.return_value = value_as_json_object
        return mock_individual

    def then_result_is(self, expected):
        self.assertEqual(expected, self._actual_result)

    def test_fact_to_json_object_for_value_without_grammar_entry(self):
        self.given_created_http_formatter()
        self.when_call_fact_to_json_object(
            self.mock_predicate_proposition(
                predicate_name="mock_predicate",
                sort_name="mock_sort",
                value="mock_value",
                value_as_json_object={"value": "mock_value"}
            ), {}
        )
        self.then_result_is({
            "sort": "mock_sort",
            "value": "mock_value",
            "grammar_entry": None,
            'perception_confidence': None,
            'understanding_confidence': None,
            'weighted_confidence': None,
            'weighted_understanding_confidence': None
        })

    def test_fact_to_json_object_for_none(self):
        self.given_created_http_formatter()
        self.when_call_fact_to_json_object(None, {})
        self.then_result_is(None)

    def test_facts_to_json_object_base_case(self):
        self.given_created_http_formatter()
        self.when_call_facts_to_json_object([
            self.mock_predicate_proposition(
                predicate_name="mock_predicate",
                sort_name="mock_sort",
                value="mock_value",
                value_as_json_object={"value": "mock_value"}
            )
        ], {
            "session_id":
            "blah",
            "entities": [{
                "sort": "mock_sort",
                "natural_language_form": "mock_grammar_entry",
                "name": "mock_value"
            }]
        })
        self.then_result_is({
            "mock_predicate": {
                "sort": "mock_sort",
                "value": "mock_value",
                "grammar_entry": "mock_grammar_entry",
                'perception_confidence': None,
                'understanding_confidence': None,
                'weighted_confidence': None,
                'weighted_understanding_confidence': None
            }
        })

    def when_call_facts_to_json_object(self, facts, session):
        self._actual_result = self._http_formatter.facts_to_json_object(facts, session)

    def test_facts_to_json_object_excludes_negative_predicate_propositions(self):
        self.given_created_http_formatter()
        self.when_call_facts_to_json_object([self.mock_predicate_proposition(polarity=Polarity.NEG)], {})
        self.then_result_is({})

    def test_facts_to_json_object_excludes_non_predicate_propositions(self):
        self.given_created_http_formatter()
        self.when_call_facts_to_json_object([self.mock_non_predicate_proposition()], {})
        self.then_result_is({})

    def test_facts_to_json_object_excludes_nullary_predicate_propositions(self):
        self.given_created_http_formatter()
        self.when_call_facts_to_json_object([self.mock_nullary_predicate_proposition()], {})
        self.then_result_is({})

    def mock_non_predicate_proposition(self, polarity=Polarity.POS):
        mock_proposition = Mock(spec=Proposition)
        mock_proposition.is_predicate_proposition.return_value = False
        mock_proposition.get_polarity.return_value = polarity
        return mock_proposition

    def mock_nullary_predicate_proposition(self, predicate_name="mock_yn_predicate", sort_name="boolean"):
        sort = self._mock_sort(sort_name)
        predicate = self._mock_predicate(predicate_name, sort)
        mock_proposition = Mock(spec=PredicateProposition)
        mock_proposition.getPredicate.return_value = predicate
        mock_proposition.getArgument.return_value = None
        return mock_proposition



from tala.model.common import Modality
from tala.model.user_move import UserMove  # noqa: F401
from tala.utils.as_json import AsJSONMixin
from tala.utils.equality import EqualityMixin
from tala.utils.unicodify import unicodify


class UnexpectedModalityException(Exception):
    pass


class Interpretation(EqualityMixin, AsJSONMixin):
    def __init__(self, moves, modality, utterance=None, perception_confidence=None):
        # type: ([UserMove], str, str) -> None
        self._moves = moves
        if modality not in Modality.SUPPORTED_MODALITIES:
            raise UnexpectedModalityException(
                f"Expected one of the supported modalities {Modality.SUPPORTED_MODALITIES} but got '{modality}'"
            )
        if utterance:
            if modality not in Modality.ALLOWS_UTTERANCE:
                raise UnexpectedModalityException(
                    f"Expected no utterance for modality '{modality}' but got '{utterance}'"
                )
        if not utterance:
            if modality in Modality.REQUIRES_UTTERANCE:
                raise UnexpectedModalityException(f"Expected an utterance for modality '{modality}' but it was missing")
        self._modality = modality
        self._utterance = utterance
        self._perception_confidence = perception_confidence

    @property
    def moves(self):
        # type: () -> [UserMove]
        return self._moves

    @property
    def modality(self):
        # type: () -> str
        return self._modality

    @property
    def utterance(self):
        # type: () -> str
        return self._utterance

    @property
    def perception_confidence(self):
        return self._perception_confidence

    def as_dict(self):
        return {
            "modality": self.modality,
            "moves": [move.as_dict() for move in self.moves],
            "utterance": self.utterance,
        }

    def __repr__(self):
        return f"{self.__class__.__name__}({unicodify(self._moves)}, {self._modality}, {self._utterance}, {self._perception_confidence})"


class InterpretationWithoutUtterance(Interpretation):
    def __init__(self, moves, modality):
        # type: ([UserMove], str) -> None
        super(InterpretationWithoutUtterance, self).__init__(moves, modality)
        self._moves = moves
        self._modality = modality

    def as_dict(self):
        return {"modality": self.modality, "moves": [move.as_dict() for move in self.moves]}

    def __repr__(self):
        return f"{self.__class__.__name__}({unicodify(self._moves)}, {self._modality})"



import unittest

from mock import patch

from tala.utils.chdir import chdir
from tala.config import BackendConfig, OverriddenDddConfig
from tala.ddd.loading import ddd_set_loader
from tala.ddd.loading.ddd_set_loader import DddNotFoundException, \
    OverriddenDddConfigNotFoundException
from tala.ddd.ddd_component_manager import DDDComponentManager
from tala.log import logger
from tala.ddd.ddd_specific_components import DDDSpecificComponents
from tala.ddd.loading import component_set_loader
from tala.ddd.loading.component_set_loader import ComponentSetLoader


class TestComponentSetLoader(unittest.TestCase):
    def setUp(self):
        self._ddd_component_manager = DDDComponentManager()
        self._overridden_ddd_config_paths = None
        self.test_logger = logger.configure_and_get_test_logger()
        self._ddds = []
        self._result = None
        self._MockDddLoader = None

    def test_ensure_ddds_loaded(self):
        with chdir("ddds"):
            self._given_component_set_loader_created()
            self._when_ensuring_ddds_loaded_from("mockup_travel.json")
            self._then_components_are_loaded_for("mockup_travel")

    def _given_component_set_loader_created(self):
        self._component_set_loader = ComponentSetLoader(self._ddd_component_manager, self._overridden_ddd_config_paths)

    def _when_ensuring_ddds_loaded_from(self, config_path):
        config = BackendConfig(config_path).read()
        ddds = config["ddds"]
        self._component_set_loader.ensure_ddds_loaded(ddds, rerank_amount=config["rerank_amount"])

    def _then_components_are_loaded_for(self, ddd_name):
        components = self._ddd_component_manager.get_ddd_specific_components(ddd_name)
        self.assertTrue(isinstance(components, DDDSpecificComponents))

    @patch("{}.DddConfig".format(ddd_set_loader.__name__))
    @patch("{}.ComponentLoader".format(component_set_loader.__name__))
    def test_ddd_config_override(self, MockDddLoader, MockDddConfig):
        with chdir("ddds"):
            self._given_overridden_ddd_config_paths([OverriddenDddConfig("hello_world", "mock_config.json")])
            self._given_component_set_loader_created()
            self._when_ensuring_ddds_loaded_from("hello_world.json")
            self._then_method_called_with(MockDddConfig, "mock_config.json")

    def _given_overridden_ddd_config_paths(self, overridden_ddd_config_paths):
        self._overridden_ddd_config_paths = overridden_ddd_config_paths

    def _then_method_called_with(self, method, argument):
        method.assert_called_with(argument)

    @patch("{}.DddConfig".format(ddd_set_loader.__name__))
    def test_non_existing_ddd_in_ddd_config_override_raises_exception(self, MockDddConfig):
        with chdir("ddds"):
            self._given_overridden_ddd_config_paths([OverriddenDddConfig("non_existing_ddd", "mock_config.json")])
            self._when_creating_ddd_set_loader_then_exception_is_raised_with_message_that_matches_regex(
                DddNotFoundException,
                "Expected overridden DDD 'non_existing_ddd' to exist in the working directory '.*', but it doesn't."
            )

    def _when_creating_ddd_set_loader_then_exception_is_raised_with_message_that_matches_regex(
        self, exception_class, expected_regex
    ):
        with self.assertRaisesRegex(exception_class, expected_regex):
            ComponentSetLoader(self._ddd_component_manager, self._overridden_ddd_config_paths)

    def test_non_existing_ddd_config_in_ddd_config_override_raises_exception(self):
        with chdir("ddds"):
            self._given_overridden_ddd_config_paths([OverriddenDddConfig("mockup_travel", "non_existing_config.json")])
            self._when_creating_ddd_set_loader_then_exception_is_raised_with_message_that_matches_regex(
                OverriddenDddConfigNotFoundException,
                "Expected DDD config '.*non_existing_config.json' to exist in DDD 'mockup_travel' but it was not found."
            )

    def _given_mock_ddd_loader(self, MockDddLoader):
        self._MockDddLoader = MockDddLoader

    def _given_ddds(self, *ddds):
        self._ddds = ddds

    def _when_loading_ddds_then_exception_is_raised_with_message_that_matches(self, ExpectedException, expected_regex):
        with self.assertRaisesRegex(ExpectedException, expected_regex):
            self._component_set_loader.ensure_ddds_loaded(
                self._ddds, logger=self.test_logger, rerank_amount=BackendConfig.DEFAULT_RERANK_AMOUNT
            )



import unittest

from mock import Mock

from tala.model.ontology import Ontology
from tala.model.person_name import PersonName

from tala.ddd.services.parameters.retriever import ParameterRetriever, ParameterNotFoundException
from tala.ddd.services.parameters.binding import SingleInstanceParameterBinding, MultiInstanceParameterBinding
from tala.ddd.services.service_interface import ServiceInterface, ServiceActionInterface, DeviceModuleTarget, \
    ServiceParameter, ParameterField
from tala.model.individual import Individual
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition
from tala.model.set import Set
from tala.model.polarity import Polarity
from tala.model.predicate import Predicate
from tala.model.proposition import PredicateProposition, ResolvednessProposition
from tala.model.sort import CustomSort, StringSort, IntegerSort, PersonNameSort, BuiltinSort


class ParameterRetrieverTestCase(unittest.TestCase):
    def setUp(self):
        self._mocked_service_interface = None
        self._parameter_retriever = None
        self._ontology = None

    def _given_ontology(self, name="mock_ontology", sorts=None, predicates=None, individuals=None, actions=None):
        if sorts is None:
            sorts = set()
        if predicates is None:
            predicates = set()
        if individuals is None:
            individuals = {}
        if actions is None:
            actions = set()

        self._ontology = Ontology(name, sorts, predicates, individuals, actions)

    def _custom_proposition(self, predicate, individual, sort_name, polarity=Polarity.POS, multiple_instances=False):
        sort = self._sort(sort_name)
        return self._proposition(predicate, individual, sort, polarity=polarity, multiple_instances=multiple_instances)

    def _proposition(self, predicate, individual, sort, polarity=Polarity.POS, multiple_instances=False):
        return PredicateProposition(
            self._predicate(predicate, sort, multiple_instances), self._individual(individual, sort), polarity=polarity
        )

    def _predicate(self, name, sort, multiple_instances=False):
        return Predicate("mocked_ontology", name, sort, multiple_instances=multiple_instances)

    def _individual(self, name, sort):
        return Individual("mocked_ontology", name, sort)

    def _sort(self, name, is_dynamic=False):
        return CustomSort("mocked_ontology", name, is_dynamic)

    def _lambda_abstraction(self, predicate, sort_name):
        sort = self._sort(sort_name)
        return LambdaAbstractedPredicateProposition(self._predicate(predicate, sort), "mocked_ontology")

    def _given_service_interface(self, action):
        self._mocked_service_interface = Mock(spec=ServiceInterface)
        self._mocked_service_interface.get_action.return_value = action

    def _given_created_parameter_retriever(self):
        self._parameter_retriever = ParameterRetriever(self._mocked_service_interface, self._ontology)

    def _then_result_is(self, expected_result):
        self.assertEqual(self._result, expected_result)

    def test_preprocess_arguments_base_case(self):
        self._given_ontology(
            sorts=set([self._sort("city")]), predicates=set([self._predicate("dest_city", self._sort("city"))])
        )
        self._given_service_interface(
            action=ServiceActionInterface(
                "MakeReservation", DeviceModuleTarget(device="MockedDevice"), [ServiceParameter("dest_city")], []
            )
        )
        self._given_created_parameter_retriever()
        self._when_preprocessing_action_arguments(
            name="MakeReservation", facts=Set([
                self._custom_proposition("dest_city", "paris", "city"),
            ])
        )
        self._then_result_is([
            SingleInstanceParameterBinding(
                ServiceParameter("dest_city"), "paris", self._custom_proposition("dest_city", "paris", "city")
            )
        ])

    def test_non_predicate_proposition_in_facts_is_tolerated(self):
        self._given_ontology(
            sorts=set([self._sort("city")]), predicates=set([self._predicate("dest_city", self._sort("city"))])
        )
        self._given_service_interface(
            action=ServiceActionInterface(
                "MakeReservation", DeviceModuleTarget(device="MockedDevice"), [ServiceParameter("dest_city")], []
            )
        )
        self._given_created_parameter_retriever()
        self._when_preprocessing_action_arguments(
            name="MakeReservation",
            facts=Set([
                self._custom_proposition("dest_city", "paris", "city"),
                ResolvednessProposition(self._lambda_abstraction("dest_country", "country")),
            ])
        )
        self._then_result_is([
            SingleInstanceParameterBinding(
                ServiceParameter("dest_city"), "paris", self._custom_proposition("dest_city", "paris", "city")
            )
        ])

    def _when_preprocessing_action_arguments(self, name, facts):
        self._result = self._parameter_retriever.preprocess_action_arguments(name, facts, {})

    def _then_result_has_values(self, expected_values):
        actual_values = [binding.value for binding in self._result]
        self.assertEqual(expected_values, actual_values)

    def test_negative_propositions_in_facts_are_ignored(self):
        self._given_ontology(
            sorts=set([self._sort("city")]), predicates=set([self._predicate("dest_city", self._sort("city"))])
        )
        self._given_service_interface(
            action=ServiceActionInterface(
                "MakeReservation", DeviceModuleTarget(device="MockedDevice"), [ServiceParameter("dest_city")], []
            )
        )
        self._given_created_parameter_retriever()
        self._when_preprocessing_action_arguments_then_exception_is_raised(
            name="MakeReservation",
            facts=Set([self._custom_proposition("dest_city", "paris", "city", polarity=Polarity.NEG)]),
            expected_exception=ParameterNotFoundException,
            expected_message="failed to get argument 'dest_city'"
        )

    def test_exception_raised_for_missing_required_parameter(self):
        self._given_ontology(
            sorts=set([self._sort("city")]), predicates=set([self._predicate("dest_city", self._sort("city"))])
        )
        self._given_service_interface(
            action=ServiceActionInterface(
                "MakeReservation", DeviceModuleTarget(device="MockedDevice"), [ServiceParameter("dest_city")], []
            )
        )
        self._given_created_parameter_retriever()
        self._when_preprocessing_action_arguments_then_exception_is_raised(
            name="MakeReservation",
            facts=Set([]),
            expected_exception=ParameterNotFoundException,
            expected_message="failed to get argument 'dest_city'"
        )

    def _when_preprocessing_action_arguments_then_exception_is_raised(
        self, name, facts, expected_exception, expected_message
    ):
        with self.assertRaisesRegex(expected_exception, expected_message):
            self._parameter_retriever.preprocess_action_arguments(name, facts, {})

    def test_multiple_instance_argument_bound_to_all_values_and_propositions(self):
        self._given_ontology(
            sorts=set([self._sort("city")]),
            predicates=set([self._predicate("dest_city", self._sort("city"), multiple_instances=True)])
        )
        self._given_service_interface(
            action=ServiceActionInterface(
                "MakeReservation", DeviceModuleTarget(device="MockedDevice"), [ServiceParameter("dest_city")], []
            )
        )
        self._given_created_parameter_retriever()
        self._when_preprocessing_action_arguments(
            name="MakeReservation",
            facts=Set([
                self._custom_proposition("dest_city", "paris", "city", multiple_instances=True),
                self._custom_proposition("dest_city", "london", "city", multiple_instances=True),
            ])
        )
        self._then_result_is([
            MultiInstanceParameterBinding(
                ServiceParameter("dest_city", is_optional=False),
                values=["paris", "london"],
                propositions=[
                    self._custom_proposition("dest_city", "paris", "city", multiple_instances=True),
                    self._custom_proposition("dest_city", "london", "city", multiple_instances=True)
                ]
            )
        ])

    def _given_an_individual_of_builtin_sort(self, value):
        self._mock_builtin_sort = Mock(spec=BuiltinSort)
        self._mock_builtin_sort.is_builtin.return_value = True
        self._individual_of_builtin_sort = Mock(spec=Individual)
        self._individual_of_builtin_sort.getSort.return_value = self._mock_builtin_sort
        self._individual_of_builtin_sort.getValue.return_value = value

    def _fact(self, predicate_name, individual, polarity=Polarity.POS, multiple_instances=False):
        sort = individual.getSort()
        return PredicateProposition(
            self._predicate(predicate_name, sort, multiple_instances), individual, polarity=polarity
        )

    def _then_generator_for_individual_of_builtin_sort_is_called_with(
        self, MockGeneratorForIndividualOfBuiltinSort, *args
    ):
        MockGeneratorForIndividualOfBuiltinSort.generate.assert_called_once_with(*args)

    def test_string_parameter(self):
        self._given_ontology(
            sorts=set([self._sort("city")]), predicates=set([self._predicate("dest_city", self._sort("city"))])
        )
        self._given_service_interface(
            action=ServiceActionInterface(
                "MakeReservation", DeviceModuleTarget(device="MockedDevice"),
                [ServiceParameter("dest_city", format=ParameterField.VALUE)], []
            )
        )
        self._given_created_parameter_retriever()
        self._when_preprocessing_action_arguments(
            name="MakeReservation", facts=Set([
                self._proposition("dest_city", "Paris", StringSort()),
            ])
        )
        self._then_result_has_values(["Paris"])

    def test_integer_parameter(self):
        self._given_ontology(predicates=set([self._predicate("n_stops", IntegerSort())]))
        self._given_service_interface(
            action=ServiceActionInterface(
                "MakeReservation", DeviceModuleTarget(device="MockedDevice"), [ServiceParameter("n_stops")], []
            )
        )
        self._given_created_parameter_retriever()
        self._when_preprocessing_action_arguments(
            name="MakeReservation", facts=Set([
                self._proposition("n_stops", 3, IntegerSort()),
            ])
        )
        self._then_result_has_values([3])

    def test_value_of_person_name_parameter(self):
        self._given_ontology(predicates=set([self._predicate("name_of_contact_to_call", PersonNameSort())]))
        self._given_service_interface(
            action=ServiceActionInterface(
                "Call", DeviceModuleTarget(device="MockedDevice"),
                [ServiceParameter("name_of_contact_to_call", ParameterField.VALUE)], []
            )
        )
        self._given_created_parameter_retriever()
        self._when_preprocessing_action_arguments(
            name="Call",
            facts=Set([
                self._proposition("name_of_contact_to_call", PersonName("Anna"), PersonNameSort()),
            ])
        )
        self._then_result_has_values(["Anna"])



# flake8: noqa

from tala.model.move import ICMMove, ICMMoveWithSemanticContent
from tala.model import plan_item
from tala.model.proposition import ServiceResultProposition
from tala.testing.lib_test_case import LibTestCase
from tala.utils.unicodify import unicodify


class PlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.question = self.dest_city_question
        self.respond_plan_item = plan_item.Respond(self.question)
        self.findout_plan_item = plan_item.Findout(self.domain_name, self.question)
        self.raise_plan_item = plan_item.Raise(self.domain_name, self.question)
        self.bind_plan_item = plan_item.Bind(self.question)
        self.if_condition = self.proposition_dest_city_paris
        self.if_consequent = [self.findout_plan_item]
        self.if_alternative = [self.raise_plan_item]
        self.if_then_else_plan_item = plan_item.IfThenElse(self.if_condition, self.if_consequent, self.if_alternative)
        self.service_result_proposition = ServiceResultProposition(
            self.ontology_name, "mockup_service_action", [], "mock_perform_result"
        )
        self.service_report_item = plan_item.ServiceReport(self.service_result_proposition)

    def test_equality_with_none(self):
        self.assertNotEqual(self.respond_plan_item, None)

    def test_create_DoPlanItem(self):
        do_item = plan_item.Do(self.buy_action)
        self.assertEqual(self.buy_action, do_item.content)

    def test_DoPlanItem_type(self):
        do_item = plan_item.Do(self.buy_action)
        self.assertEqual(plan_item.TYPE_DO, do_item.type_)

    def test_isRespondPlanItem(self):
        self.assertTrue(self.respond_plan_item.type_ == plan_item.TYPE_RESPOND)

    def test_isFindoutPlanItem(self):
        self.assertTrue(self.findout_plan_item.type_ == plan_item.TYPE_FINDOUT)

    def test_FindoutPlanItem_content(self):
        self.assertEqual(self.question, self.findout_plan_item.content)

    def test_isRaisePlanItem(self):
        self.assertTrue(self.raise_plan_item.type_ == plan_item.TYPE_RAISE)

    def test_RaisePlanItem_content(self):
        self.assertEqual(self.question, self.raise_plan_item.content)

    def test_FindoutPlanItem_to_string(self):
        expected_string = "findout(?X.dest_city(X))"
        self.assertEqual(expected_string, str(self.findout_plan_item))

    def test_PlanItem_unicode(self):
        quitPlanItem = plan_item.Quit()
        self.assertEqual(str(quitPlanItem), "quit")

    def test_GreetPlanItem_unicode(self):
        greetPlanItem = plan_item.Greet()
        self.assertEqual(str(greetPlanItem), "greet")

    def test_findout_is_question_plan_item(self):
        self.assertTrue(self.findout_plan_item.type_ in plan_item.QUESTION_TYPES)

    def test_raise_is_question_plan_item(self):
        self.assertTrue(self.raise_plan_item.type_ in plan_item.QUESTION_TYPES)

    def test_bind_is_question_plan_item(self):
        self.assertTrue(self.bind_plan_item.type_ in plan_item.QUESTION_TYPES)

    def test_respond_is_not_question_plan_item(self):
        self.assertFalse(self.respond_plan_item.type_ in plan_item.QUESTION_TYPES)

    def test_findout_is_question_raising_item(self):
        self.assertTrue(self.findout_plan_item.is_question_raising_item())

    def test_raise_is_question_raising_item(self):
        self.assertTrue(self.raise_plan_item.is_question_raising_item())

    def test_bind_is_not_question_raising_item(self):
        self.assertFalse(self.bind_plan_item.is_question_raising_item())

    def test_respond_is_not_question_raising_item(self):
        self.assertFalse(self.respond_plan_item.is_question_raising_item())

    def test_emit_item_for_icm_und_non_pos_is_question_raising_item(self):
        icm_und = ICMMove(ICMMove.UND, polarity=ICMMove.NEG)
        emit_item = plan_item.EmitIcm(icm_und)
        self.assertTrue(emit_item.is_question_raising_item())

    def test_emit_item_for_icm_und_pos_for_positive_content_is_question_raising_item(self):
        positive_content = self.proposition_dest_city_paris
        icm = ICMMoveWithSemanticContent(ICMMove.UND, positive_content, polarity=ICMMove.POS)
        emit_item = plan_item.EmitIcm(icm)
        self.assertTrue(emit_item.is_question_raising_item())

    def test_emit_item_for_icm_und_pos_for_negated_content_is_not_question_raising_item(self):
        negation = self.proposition_not_dest_city_paris
        icm = ICMMoveWithSemanticContent(ICMMove.UND, negation, polarity=ICMMove.POS)
        emit_item = plan_item.EmitIcm(icm)
        self.assertFalse(emit_item.is_question_raising_item())

    def test_report_item_is_turn_yielding(self):
        self.assertTrue(self.service_report_item.is_turn_yielding())

    def test_respond_item_is_turn_yielding(self):
        self.assertTrue(self.respond_plan_item.is_turn_yielding())

    def test_emit_item_for_acc_neg_is_turn_yielding(self):
        icm_acc_neg = ICMMove(ICMMove.ACC, polarity=ICMMove.NEG)
        emit_item = plan_item.EmitIcm(icm_acc_neg)
        self.assertTrue(emit_item.is_turn_yielding())

    def test_emit_item_for_acc_pos_is_not_turn_yielding(self):
        icm_acc_pos = ICMMove(ICMMove.ACC, polarity=ICMMove.POS)
        emit_item = plan_item.EmitIcm(icm_acc_pos)
        self.assertFalse(emit_item.is_turn_yielding())

    def test_if_then_else_getters(self):
        self.assertEqual(plan_item.TYPE_IF_THEN_ELSE, self.if_then_else_plan_item.type_)
        self.assertEqual(self.if_condition, self.if_then_else_plan_item.get_condition())
        self.assertEqual(self.if_consequent, self.if_then_else_plan_item.get_consequent())
        self.assertEqual(self.if_alternative, self.if_then_else_plan_item.get_alternative())

    def test_if_then_else_to_string(self):
        expected_string = "if_then_else{}".format(
            unicodify((self.if_condition, self.if_consequent, self.if_alternative))
        )
        self.assertEqual(expected_string, str(self.if_then_else_plan_item))

    def test_create_forget_all(self):
        item = plan_item.ForgetAll()
        self.assertTrue(item.type_ == plan_item.TYPE_FORGET_ALL)

    def test_create_forget_plan_item_proposition(self):
        fact = self.proposition_dest_city_paris
        item = plan_item.Forget(fact)
        self.assertEqual(item.type_, plan_item.TYPE_FORGET)
        self.assertEqual(fact, item.content)

    def test_create_forget_plan_item_predicate(self):
        predicate = self.predicate_dept_city
        item = plan_item.Forget(predicate)
        self.assertTrue(item.type_ == plan_item.TYPE_FORGET)
        self.assertEqual(predicate, item.content)

    def test_create_forget_issue_plan_item(self):
        issue = self.price_question
        item = plan_item.ForgetIssue(issue)
        self.assertTrue(item.type_ == plan_item.TYPE_FORGET_ISSUE)
        self.assertEqual(issue, item.content)

    def test_service_report_plan_item(self):
        self.assertTrue(self.service_report_item.type_ == plan_item.TYPE_SERVICE_REPORT)
        self.assertTrue(self.service_result_proposition, self.service_report_item.content)

    def test_findout_equality(self):  #/// Should allow_answer_from_pcom change equality?
        item = plan_item.Findout(self.domain_name, self.question)
        identical_item = plan_item.Findout(self.domain_name, self.question)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(identical_item, item)

    def test_inequality_due_to_question(self):
        item1 = plan_item.Findout(self.domain_name, self.dest_city_question)
        item2 = plan_item.Findout(self.domain_name, self.price_question)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(item1, item2)

    def test_clone_findout_into_raise_plan_item(self):
        findout_plan_item = plan_item.Findout(self.domain_name, self.question)
        raise_plan_item = findout_plan_item.clone_as_type(plan_item.TYPE_RAISE)
        self.assertTrue(raise_plan_item.type_ == plan_item.TYPE_RAISE)
        self.assertEqual(self.question, raise_plan_item.question)

    def test_findout_defaults_to_disallow_answer_from_pcom(self):
        findout_plan_item = plan_item.Findout(self.domain_name, self.question)
        self.assertFalse(findout_plan_item.allow_answer_from_pcom)

    def test_findout_supports_allow_answer_from_pcom(self):
        findout_plan_item = plan_item.Findout(self.domain_name, self.question, allow_answer_from_pcom=True)
        self.assertTrue(findout_plan_item.allow_answer_from_pcom)

    #/// FindoutPlanItem is not used by ddd_xml_compiler. Only
    # QuestionRaisingPlanItem is.
    # 1. Tests?
    # 2. Remove FindoutPlanItem?


class InvokeServiceActionPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_is_invoke_service_action_plan_item(self):
        self.given_created_service_action_plan_item()
        self._actual_result = self._plan_item.type_
        self.then_result_is(plan_item.TYPE_INVOKE_SERVICE_ACTION)

    def given_created_service_action_plan_item(self, service_action="mock_service_action", **kwargs):
        self._plan_item = plan_item.InvokeServiceAction(self.ontology.name, service_action, **kwargs)

    def test_get_service_action(self):
        self.given_created_service_action_plan_item(service_action="mock_service_action")
        self.when_call(self._plan_item.get_service_action)
        self.then_result_is("mock_service_action")

    def test_has_interrogative_preconfirmation_false_by_default(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.has_interrogative_preconfirmation)
        self.then_result_is(False)

    def test_interrogative_preconfirmation_true(self):
        self.given_created_service_action_plan_item(preconfirm=plan_item.InvokeServiceAction.INTERROGATIVE)
        self.when_call(self._plan_item.has_interrogative_preconfirmation)
        self.then_result_is(True)

    def test_has_assertive_preconfirmation_false_by_default(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.has_assertive_preconfirmation)
        self.then_result_is(False)

    def test_assertive_preconfirmation_true(self):
        self.given_created_service_action_plan_item(preconfirm=plan_item.InvokeServiceAction.ASSERTIVE)
        self.when_call(self._plan_item.has_assertive_preconfirmation)
        self.then_result_is(True)

    def test_has_postconfirmation_false_by_default(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.has_postconfirmation)
        self.then_result_is(False)

    def test_has_postconfirmation_true(self):
        self.given_created_service_action_plan_item(postconfirm=True)
        self.when_call(self._plan_item.has_postconfirmation)
        self.then_result_is(True)

    def test_should_downdate_plan_true_by_default(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.should_downdate_plan)
        self.then_result_is(True)

    def test_should_downdate_plan_false(self):
        self.given_created_service_action_plan_item(downdate_plan=False)
        self.when_call(self._plan_item.should_downdate_plan)
        self.then_result_is(False)

    def test_unicode(self):
        self.given_created_service_action_plan_item(
            service_action="mock_service_action", preconfirm=None, postconfirm=False, downdate_plan=True
        )
        self.when_get_unicode()
        self.then_result_is(
            "invoke_service_action(mock_service_action, {preconfirm=None, postconfirm=False, downdate_plan=True})"
        )

    def when_get_unicode(self):
        self._actual_result = str(self._plan_item)


class InvokeServiceQueryPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_is_invoke_service_query_plan_item(self):
        self.given_created_invoke_service_query_plan_item()
        self._actual_result = self._plan_item.type_
        self.then_result_is(plan_item.TYPE_INVOKE_SERVICE_QUERY)

    def given_created_invoke_service_query_plan_item(self, *args, **kwargs):
        self._create_invoke_service_query_plan_item(*args, **kwargs)

    def _create_invoke_service_query_plan_item(self, issue="mock_issue", min_results=None, max_results=None):
        self._plan_item = plan_item.InvokeServiceQuery(issue, min_results, max_results)

    def test_min_results_is_0_by_default(self):
        self.given_created_invoke_service_query_plan_item()
        self.when_call(self._plan_item.get_min_results)
        self.then_result_is(0)

    def test_min_results_overridden(self):
        self.given_created_invoke_service_query_plan_item(min_results=1)
        self.when_call(self._plan_item.get_min_results)
        self.then_result_is(1)

    def test_max_results_is_none_by_default(self):
        self.given_created_invoke_service_query_plan_item()
        self.when_call(self._plan_item.get_max_results)
        self.then_result_is(None)

    def test_max_results_overridden(self):
        self.given_created_invoke_service_query_plan_item(max_results=1)
        self.when_call(self._plan_item.get_max_results)
        self.then_result_is(1)

    def test_exception_raised_for_min_results_below_0(self):
        self.when_created_invoke_service_query_plan_item_then_exception_is_raised(
            min_results=-1,
            max_results=None,
            expected_exception=plan_item.MinResultsNotSupportedException,
            expected_message="Expected 'min_results' to be 0 or above but got -1."
        )

    def test_exception_raised_for_max_results_below_1(self):
        self.when_created_invoke_service_query_plan_item_then_exception_is_raised(
            min_results=0,
            max_results=0,
            expected_exception=plan_item.MaxResultsNotSupportedException,
            expected_message="Expected 'max_results' to be None or above 0 but got 0."
        )

    def when_created_invoke_service_query_plan_item_then_exception_is_raised(
        self, min_results, max_results, expected_exception, expected_message
    ):
        with self.assertRaises(expected_exception) as context_manager:
            self._create_invoke_service_query_plan_item(min_results=min_results, max_results=max_results)
        self.assertEqual(expected_message, str(context_manager.exception))

    def test_get_content(self):
        self.given_created_invoke_service_query_plan_item(issue="mock_issue")
        self._actual_result = self._plan_item.content
        self.then_result_is("mock_issue")

    def test_unicode(self):
        self.given_created_invoke_service_query_plan_item(issue="mock_issue", min_results=0, max_results=1)
        self.when_get_unicode()
        self.then_result_is("invoke_service_query(mock_issue, min_results=0, max_results=1)")

    def when_get_unicode(self):
        self._actual_result = str(self._plan_item)


class JumpToPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.action = self.buy_action
        self.item = plan_item.JumpTo(self.action)

    def test_unicode(self):
        self.assertEqual("jumpto(buy)", str(self.item))

    def test_equality(self):
        identical_item = plan_item.JumpTo(self.action)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.item, identical_item)

    def test_inequality(self):
        non_identical_item = plan_item.JumpTo(self.top_action)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.item, non_identical_item)


class AssumePlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.assume_price = plan_item.Assume(self.price_proposition)

    def test_unicode(self):
        self.assertEqual("assume(price(1234.0))", str(self.assume_price))

    def test_is_assume_item(self):
        self.assertEqual(self.assume_price.type_, plan_item.TYPE_ASSUME)


class AssumeSharedPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.assume_shared_price = plan_item.AssumeShared(self.price_proposition)

    def test_unicode(self):
        self.assertEqual("assume_shared(price(1234.0))", str(self.assume_shared_price))

    def test_is_assume_shared_item(self):
        self.assertEqual(self.assume_shared_price.type_, plan_item.TYPE_ASSUME_SHARED)


class AssumeSharedIssuePlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.assume_price_issue = plan_item.AssumeIssue(self.price_question)

    def test_unicode(self):
        self.assertEqual("assume_issue(?X.price(X))", str(self.assume_price_issue))

    def test_is_assume_shared_issue(self):
        self.assertEqual(self.assume_price_issue.type_, plan_item.TYPE_ASSUME_ISSUE)


class LogPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.message = "log message"
        self.log_plan_item = plan_item.Log(self.message)

    def test_string(self):
        self.assertEqual("log_plan_item('log message')", str(self.log_plan_item))

    def test_is_log_plan_item(self):
        self.assertEqual(self.log_plan_item.type_, plan_item.TYPE_LOG)
